from financepype.operations.proposal import OperationProposal


class OrderProposal(OperationProposal):
    """A class representing a proposal for an order.

    This class contains information about a proposal for an order,
    including the order details and the associated operation.
    """

    pass
